import java.io.*;
import java.text.*;
import java.util.*;

/**
 * Pascal-2001 Scanner<BR>
 * ICS 142<BR>
 * Justin Erenkrantz<BR>
 * 10/13/2000<BR>
 * 
 * This class will attempt to scan a Pascal-2001 file.
 * The filename is read in from the command line.
 * Output goes to a file called pas.out.
 *
 */
public class PascalScanner
{
   private final String EOT = "<EOT>";
   private final String OUTPUT_FILE = "pas.out";

   private InputStream inputStream;
   private StringBuffer currentStringBuffer;
   private StringBuffer outputStringBuffer;

   private List wordOps;
   private List termOps;
   private List nonTermOps;
   private List allOps;

   private char currentChar;
   private short currentType;

   // Both of these are accessible to outside classes
   public int currentToken;
   public String currentTokenValue;

   private boolean isClosed;
   private int isInCommentBlock;

   public PascalScanner()
   {
      currentStringBuffer = null;

      isInCommentBlock = 0;

      wordOps = setArray(PascalLanguage.reservedWords, PascalLanguage.keywords);
      termOps = setArray(PascalLanguage.terminalOps, PascalLanguage.keywords);
      nonTermOps = setArray(PascalLanguage.nonTerminalOps, PascalLanguage.keywords);
      allOps = Arrays.asList(PascalLanguage.keywords);
   }

   public void setInputStream(InputStream is)
   {
      isClosed = false;
      inputStream = is;
      currentChar = getRawChar();
   }

   public void setInputStream(StringBuffer sb)
   {
      // convert to a byte[] reader...grr...
      String s = sb.toString();
      byte[] b = s.getBytes();
      ByteArrayInputStream bais = new ByteArrayInputStream(b);
      setInputStream(bais);
   }

   private boolean hasMoreTokens()   
   {
      return !isClosed;
   }

   private List setArray(int[] keys, Object[] values)
   {
      Vector v = new Vector(values.length);
      
      for (int i = 0; i < keys.length; i++)
         v.add(values[keys[i]]);
      
      return v;
   }

   private boolean isOperator(String s)
   {
      return (termOps.contains(s) || nonTermOps.contains(s));
   }   

   private char getRawChar()
   {
      // on error, force me to be skipped...woohoo!
      char c = ' ';

      try
      {
         // it better be a char
         int b = inputStream.read();
         if (b == -1)
         {
            inputStream.close();
            isClosed = true;
         }
         else
            c = (char)b;
      }
      catch (IOException ie)
      {
         ie.printStackTrace();
      }

      //System.out.print(c + "\t");
      return c;
   }

   private char getChar()
   {
      if (hasMoreTokens())
      {
         char c = getRawChar();

         currentChar = c;

         return c;
      }
      else
         throw new StringIndexOutOfBoundsException();
   }

   private void parseDigits()
   {
      char c = currentChar;
      do
      {
         currentStringBuffer.append(c);
         c = getChar();   
      }
      while (Character.isDigit(c));
      
      currentType = PascalLanguage.SCAN_DIGIT;
   
      //System.out.println("digit: " + currentStringBuffer);
   }

   private void parseLetters()
   {
      char c = currentChar;
      do
      {
         currentStringBuffer.append(c);
         c = getChar();
      }
      while (Character.isLetterOrDigit(c));
      
      currentType = PascalLanguage.SCAN_LETTER;

      //System.out.println("letter: " + currentStringBuffer);
   }

   private void parseSpecialCharacters()
   {
      char c = currentChar;
      do
      {
         currentStringBuffer.append(c);

         if (isInCommentBlock == 0 && !allOps.contains(currentStringBuffer.toString()))
         {
            currentStringBuffer.setLength(currentStringBuffer.length() - 1);
            break;
         }

         c = getChar();
      }
      while (
            !Character.isLetterOrDigit(c) && 
            !Character.isWhitespace(c) &&
            !termOps.contains(currentStringBuffer.toString())
            );
      
      currentType = PascalLanguage.SCAN_SPECIAL;

      //System.out.println("special: " + currentStringBuffer);
   }

   public void displayToken(int token)
   {
      String currentString = currentStringBuffer.toString();
      if (token != PascalLanguage.TOKEN_INVALID)
         outputStringBuffer.append(" " + allOps.get(token));
/*
      switch (type)
      {
      case PascalLanguage.TOKEN_NUMBER:
         outputStringBuffer.append(" " + number");
         currentStringBuffer.setLength(0);
         break;
      case PascalLanguage.SCAN_LETTER:
         if (wordOps.contains(currentString))
            outputStringBuffer.append( " " + currentString);
         else
            outputStringBuffer.append(" ident");
         currentStringBuffer.setLength(0);
         break;
      case PascalLanguage.SCAN_SPECIAL:
         if (allOps.contains(currentString))
            outputStringBuffer.append(" " + currentString);
         else
            System.out.println("Unknown character " + currentString);
         currentStringBuffer.setLength(0);
         break;
      case PascalLanguage.SCAN_INVALID:
      default:
         break;
      }
*/
   }

   public int GetSym()
   {
      currentStringBuffer = new StringBuffer(10);

      currentType = PascalLanguage.SCAN_INVALID;

      char c = currentChar;

      while (hasMoreTokens())
      {
         if (Character.isDigit(c))
            parseDigits();
         else if (Character.isLetter(c))
            parseLetters();
         else if (Character.isWhitespace(c)) 
            c = getChar();
         else
            parseSpecialCharacters();

         if (currentType != PascalLanguage.SCAN_INVALID)
         {
            int terminalIndex = PascalLanguage.TOKEN_INVALID;
            currentTokenValue = currentStringBuffer.toString();

            if (currentType == PascalLanguage.SCAN_DIGIT)
               terminalIndex = PascalLanguage.TOKEN_NUMBER;
            else if ((terminalIndex = allOps.indexOf(currentTokenValue)) == -1)
               terminalIndex = PascalLanguage.TOKEN_IDENT;

            currentToken = terminalIndex;   

            switch (isInCommentBlock)
            {
            case 1:
               if (currentToken == PascalLanguage.TOKEN_COMMENT1_END)
                  isInCommentBlock = 0;

               return GetSym();
            case 2:
               if (currentToken == PascalLanguage.TOKEN_COMMENT2_END)
                  isInCommentBlock = 0;

               return GetSym();
            default:
               if (currentToken == PascalLanguage.TOKEN_COMMENT1_BEGIN)
               {
                  isInCommentBlock = 1;
                  return GetSym();
               }
               else if (currentToken == PascalLanguage.TOKEN_COMMENT2_BEGIN)
               {
                  isInCommentBlock = 2;
                  return GetSym();
               }
            }

            return terminalIndex;
         }

         // update and reset
         c = currentChar;
      }

      return PascalLanguage.TOKEN_INVALID;
   }

   public StringBuffer readFile(String fileName)
   {
      if (fileName == null || fileName.length() <= 0)
         return null;

      FileReader fr = null;

      try
      {
         fr = new FileReader(fileName);
      }
      catch (FileNotFoundException fnfe)
      {   
         fnfe.printStackTrace();   
         System.out.println(fileName + "not found");
         return null;
      }

      BufferedReader br = new BufferedReader(fr);   
      String currentLine = null;

      try {
         StringBuffer sb = new StringBuffer();   

         while ((currentLine = br.readLine()) != null)
            sb.append(currentLine + " ");
         
         return sb;
         //ps.parse(sb.toString());
         //ps.parse("VAR matrix=   ARRAY   5OF   ARRAY 5 OF INTEGER; ");
         //ps.parse("matrix=   5*5MOD 5; ");
      }
      catch (IOException ie)
      {   ie.printStackTrace();   }

      return null;
   }

   public StringBuffer parseFile(StringBuffer inputTokens)
   {
      outputStringBuffer = new StringBuffer();
      setInputStream(inputTokens);

      while (hasMoreTokens())
      {
         int tokenNumber = GetSym();
         //System.out.println(tokenNumber);
         displayToken(tokenNumber);
      }

      // we are now done with this line
      if (outputStringBuffer.length() > 0 && 
         outputStringBuffer.charAt(0) == ' ')
         outputStringBuffer.deleteCharAt(0);

      return outputStringBuffer;
   }

   public void writeFile(String fileName, StringBuffer outputBuffer)
   {
      try
      {   
         FileWriter fos = new FileWriter(fileName);
         fos.write(outputBuffer.toString());
         fos.close();
      }
      catch (IOException ie)
      {   
         ie.printStackTrace();
         System.out.println("Error writing to " + fileName);
      }
   }

   public static void main(String[] args)
   {
      PascalScanner ps = new PascalScanner();

      if (args.length == 0 || args[0] == null || args[0].equals(""))
      {
         System.out.println("java PascalScanner <Input Filename>");
         return;
      }
      
      StringBuffer inputFile = ps.readFile(args[0]);
      StringBuffer outputFile = ps.parseFile(inputFile);

      // Hack only needed for this assignment
      outputFile.append(" ");
      outputFile.append("<EOT>");

      //System.out.println(outputFile);
      ps.writeFile(ps.OUTPUT_FILE, outputFile);
   }
}
