import java.util.Vector;
import java.util.ListIterator;

/**
 * An inordinately simple class which contains
 * a representation of a symbol table.
 */
public class PascalSymbolTable extends Vector
{
   private PascalSymbolTable parent = null;
   
   PascalSymbolTable()
   {
      super();
   }

   public void setParent(PascalSymbolTable pst)
   {
      parent = pst;
   }

   /**
    * Guess what.
    */ 
   public void printSymbolTable(boolean subscope)
   {
      System.out.println("---");
      ListIterator li = listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            if (sym.type != null)
               System.out.println(sym);
         }
      }

      System.out.println("---");

      if (parent != null && subscope)
         parent.printSymbolTable();
   }

   public void printSymbolTable()
   {
      printSymbolTable(false);
   }
 
   public PascalSymbol getSymbol(String s, int kind)
   {
      return getSymbol(s, kind, true);    
   }
   
   /**
    * Retrieves a named symbol with that kind from the symbol table.
    *
    * @param s   Name of symbol
    * @param kind   Kind of symbol
    * @return Associated symbol
    *
    * @see PascalLanguage
    * @see PascalSymbol
    */
   public PascalSymbol getSymbol(String s, int kind, boolean subscope)
   {
      Object o = null;
      PascalSymbol temp = null;

      ListIterator li = listIterator();
      while (li != null && li.hasNext())
      { 
         o = li.next();
         if (o instanceof PascalSymbol)
         {
            temp = (PascalSymbol)o;
            
            if (temp.identifier.equals(s) &&
                temp.kind == kind)
               return temp;
         }
      }

      if (subscope == true && parent != null)
      {
         return parent.getSymbol(s, kind);
      }
      else
         return null;
   }

   public int getSymbolTableDepth()
   {
      if (parent == null)
         return -1;

      int count = 0;

      PascalSymbolTable p = this;
      do
      {
         count++;
         p = p.parent;
      }
      while (p.parent != null);

      return count;
   }

   public int getSymbolDepth(PascalSymbol ps)
   { 
      if (parent == null)
         return -99;
      else if (contains(ps))
         return 0;
      else
         return 1 + parent.getSymbolDepth(ps);
   }

   /**
    * @return The size of the current symbol table
    */
   public int getSymbolTableSize()
   {
      int size = 0;

      ListIterator li = listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            switch (sym.kind)
            {
            case PascalLanguage.KIND_VAR:
               size += sym.type.sizeOf;
               break;
            default:
               break; 
            }
         }
      }

      return size;
   }

   /**
    * Iterates through the symbol table and allocates memory
    *
    * @param ta TargetArchitecture - if NULL, then address is
    *           exactly equivalent to space.
    */
   public int generateAddresses(TargetArchitecture ta, int offset)
   {
      int size = offset, curSize = 0;

      ListIterator li = listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            switch (sym.kind)
            {
            case PascalLanguage.KIND_VAR:
               curSize = sym.type.sizeOf;
               if (ta != null)
                  sym.address = ta.allocateChunk(curSize);
               else
                  sym.address = size;
               size += curSize;
               break;
            /* Procs don't count... */
            case PascalLanguage.KIND_PROC:
               break;
            default:
               size += sym.type.sizeOf;
               break;
            } 
         }
      }

      return size;
   }

   public int generateAddresses(TargetArchitecture ta)
   {
      return generateAddresses(ta, 0);
   }

   public int generateAddresses()
   {
      return generateAddresses(null, 0);
   }

   public void reverseAddresses()
   {
      ListIterator li = listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            switch (sym.kind)
            {
            case PascalLanguage.KIND_VAR:
               sym.address = -sym.address;
               break;
            /* Procs don't count... */
            /* Constants don't count... */
            case PascalLanguage.KIND_PROC:
            default:
               break;
            } 
         }
      }
   }
}
