import java.util.List;

/**
 * This class defines the basic types.
 * 
 * Based on Dr. Franz's structure as described in lecture.
 */
public class PascalType implements Comparable
{
   int form;            // See PascalLanguage.TYPE_
   int length;          // For arrays
   int sizeOf;          // Size of 1 element of this type
   PascalSymbolTable symbolTable;    // Records have private symbol table
   PascalType base;     // Elements in Array
   PascalSymbol link;   // Linked symbol for Array
   
   public PascalType()
   {
      form = PascalLanguage.TYPE_INVALID;
      length = PascalLanguage.TYPE_INVALID;
      sizeOf = PascalLanguage.TYPE_INVALID;
      link = null;
      base = null;
      symbolTable = null;
   }

   public PascalType(int newForm, int newSize)
   {
      form   = newForm;
      sizeOf = newSize;

      length = PascalLanguage.TYPE_INVALID;
      link = null;
      base = null;
      symbolTable = null;
   }

   public String toString()
   {
      StringBuffer sb = new StringBuffer(100);
      sb.append("Form:\t" + form);
      sb.append("\tSize:\t" + sizeOf);

      if (form == PascalLanguage.TYPE_RECORD)
         sb.append("\nSymbol Table:\t" + symbolTable + "");

      if (form == PascalLanguage.TYPE_ARRAY)
      {
         sb.append("\tArrlen:\t" + length);
         sb.append("\tLink:\t" + link);
         sb.append("\nArray Type:\t" + base);
      } 

      return sb.toString();
   }

   public int compareTo(Object o)
   {
      if (o instanceof PascalType)
      {
         PascalType t = (PascalType)o;

         // FIXME: Come up with something more elegant
         if (form < t.form)
            return -1;
         else if (form > t.form)
            return 1;

         if (length < t.length)
            return -1;
         else if (length > t.length)
            return 1;

         if (sizeOf < t.sizeOf)
            return -1;
         else if (sizeOf > t.sizeOf)
            return 1;

         if (base != null)
         {
            if (t.base == null)
               return 1;
            else
               return base.compareTo(t.base);
         }
         else if (t.base != null)
            return -1;

         // If we reached this far,
         // they are equivalent for now.
         return 0;
      }
      else
         throw new ClassCastException();
   }
}
