import java.io.Writer;
import java.io.FileWriter;
import java.io.StringWriter;
import java.io.BufferedWriter;

/**
 * Pascal-2001 Target Architecture<BR>
 * ICS 142<BR>
 * Justin Erenkrantz<BR>
 * 12/3/2000<BR>
 *
 * This class has all information about the target architecture
 *
 */
public class TargetArchitecture
{
   // I don't know... 32k sounds decent
   public static final int MAX_INST = 32000;
   public static final int MAX_INT = 65536;

   // Size of primatives
   public static final int SIZE_INTEGER = 4;
   public static final int SIZE_BOOLEAN = 4;

   // Registers of notice
   public static final int REG_RET = 31;
   public static final int REG_GLOBAL = 30;
   public static final int REG_SP = 29;
   public static final int REG_FP = 28;
   public static final int REG_ZERO = 0;
   
   // 1..27 only
   public static final int REG_COUNT = 27;
   public static final int REG_INVALID = -1; 

   /* Instruction Formats */
   public static final int FORMAT_INVALID = 0;
   public static final int FORMAT_ONE = 1;
   public static final int FORMAT_TWO = 2;
   public static final int FORMAT_THREE = 3;

   /* Register-to-register instructions */
   public static final int INST_INVALID = -1;
   public static final int INST_ADD = 0;
   public static final int INST_SUB = 1;
   public static final int INST_MUL = 2;
   public static final int INST_DIV = 3;
   public static final int INST_MOD = 4;
   public static final int INST_CMP = 5;

   public static final int INST_OR  = 8;
   public static final int INST_AND = 9;
   public static final int INST_BIC = 10;
   public static final int INST_XOR = 11;

   public static final int INST_LSH = 12;
   public static final int INST_RSH = 13;
   public static final int INST_CHK = 14;

   public static final int INST_IMM_OFFSET = 16;

   /* Register and Memory instructions have an offset */
   public static final int INST_ADDI = INST_IMM_OFFSET + INST_ADD;
   public static final int INST_SUBI = INST_IMM_OFFSET + INST_SUB;
   public static final int INST_MULI = INST_IMM_OFFSET + INST_MUL;
   public static final int INST_DIVI = INST_IMM_OFFSET + INST_DIV;
   public static final int INST_MODI = INST_IMM_OFFSET + INST_MOD;
   public static final int INST_CMPI = INST_IMM_OFFSET + INST_CMP;

   public static final int INST_ORI  = INST_IMM_OFFSET + INST_OR;
   public static final int INST_ANDI = INST_IMM_OFFSET + INST_AND;
   public static final int INST_BICI = INST_IMM_OFFSET + INST_BIC;
   public static final int INST_XORI = INST_IMM_OFFSET + INST_XOR;

   public static final int INST_LSHI = INST_IMM_OFFSET + INST_LSH;
   public static final int INST_RSHI = INST_IMM_OFFSET + INST_RSH;
   public static final int INST_CHKI = INST_IMM_OFFSET + INST_CHK;

   /* Load instructions */
   public static final int INST_LDW = 32;
   public static final int INST_LDX = 33;
   public static final int INST_POP = 34;

   /* Store instructions */
   public static final int INST_STW = 36;
   public static final int INST_STX = 37;
   public static final int INST_PSH = 38;

   /* Control instructions */
   public static final int INST_BEQ = 40;
   public static final int INST_BNE = 41;
   public static final int INST_BLT = 42;
   public static final int INST_BGE = 43;
   public static final int INST_BLE = 44;
   public static final int INST_BGT = 45;
   public static final int INST_BSR = 46;
   public static final int INST_JSR = 48;
   public static final int INST_RET = 49;

   /* Input/Output instructions */
   public static final int INST_RDD = 50;
   public static final int INST_WRD = 51;
   public static final int INST_WRH = 52;
   public static final int INST_WRL = 53;

   // Actually, it's 6 bits.  I'm lazy.
   public static final int HIGHEST_OPCODE = INST_WRL + 1; 

   /* Instruction types */

   /* F3 types */
   public static final int[] oneOpInst = 
      { INST_JSR };

   /* F2 types */
   public static final int[] twoOpInst = 
      { INST_ADD, INST_SUB, INST_MUL, INST_DIV, INST_MOD, INST_CMP, 
        INST_OR,  INST_AND, INST_BIC, INST_XOR, INST_LSH, INST_RSH, 
        INST_CHK, INST_LDX, INST_STX, INST_RET, INST_RDD, INST_WRD, 
        INST_WRH };

   /* F1 types */
   public static final int[] threeOpInst = 
      { INST_ADDI, INST_SUBI, INST_MULI, INST_DIVI, INST_MODI, 
        INST_CMPI, INST_ORI,  INST_ANDI, INST_BICI, INST_XORI, 
        INST_LSHI, INST_RSHI, INST_CHKI, INST_LDW,  INST_POP, 
        INST_STW,  INST_PSH,  INST_BEQ,  INST_BNE,  INST_BLT, 
        INST_BGE,  INST_BLE,  INST_BGT,  INST_BSR,  INST_WRL } ;

   public static final String[] mneumonicNames = { "" };

   public static final int[] comparisonTokens = {
         PascalLanguage.TOKEN_EQUALS, PascalLanguage.TOKEN_NE, 
         PascalLanguage.TOKEN_LT, PascalLanguage.TOKEN_LTE, 
         PascalLanguage.TOKEN_GT, PascalLanguage.TOKEN_GTE };

   public static final int[] comparisonInstructions = {
         INST_BEQ, INST_BNE, 
         INST_BLT, INST_BLE,
         INST_BGT, INST_BGE, };

   public static final int[] comparisonAntiInstructions = {
         INST_BNE, INST_BEQ, 
         INST_BGE, INST_BGT, 
         INST_BLE, INST_BLT };

   private int freeAddress;
   private int freeRegister;

   private boolean[] allocatedRegisters;

   private int[] opTypeIndex;

   private dlx2001 processor;

   private int[] outputInstructions;
   private int instructions;

   private String objectFilename;
   private PascalWriter debugWriter;
   private PascalWriter assemblerWriter;

   public TargetArchitecture()
   {
      createInstance();
   }

   public TargetArchitecture(String objectFilename, 
                             String assemblerFilename,
                             String debugFilename)
   {
      createInstance();
      this.objectFilename = objectFilename;
      debugWriter = new PascalWriter(debugFilename); 
      assemblerWriter = new PascalWriter(assemblerFilename); 
   }

   private void createInstance()
   {
      freeAddress = 0;
      freeRegister = 0;
      opTypeIndex = createOperatorIndexTable();
      
      // Of course, I find this on Dr. Franz's website
      // AFTER I am done with this.  *Sigh*
      // Never to late to retrofit
      processor = new dlx2001();

      outputInstructions = new int[MAX_INST];
      instructions = 0;

      allocatedRegisters = new boolean[REG_COUNT];
   }

   public int allocateChunk(int size)
   {
      int returnValue = freeAddress;
      freeAddress -= size;
      return returnValue;
   }

   public int allocateRegister()
   {
      // For our needs, a linear search is fine.
      for (int i = 0; i < allocatedRegisters.length; i++)
         if (allocatedRegisters[i] == false)
         {
            allocatedRegisters[i] = true;
            return ++i;
         }

      // no free registers.  oops.
      return -1;
   }

   public void allocateRegister(int reg)
   {
      // we return + 1, so subtract one
      allocatedRegisters[--reg] = true;
   }

   public void deallocateRegister(int reg)
   {
      // we return + 1, so subtract one
      allocatedRegisters[--reg] = false;
   }

/*
   public int allocateRegister()
   {
      return ++freeRegister;
   }

   public int getRegister()
   {
      return freeRegister;
   }   

   public int deallocateRegister()
   {
      return --freeRegister;
   }

   public int postDeallocateRegister()
   {
      return freeRegister--;
   }
*/
   public int getVarReg()
   {
      return REG_FP;
   }

/*
    private List setArray(int[] keys, Object[] values)
    {
        Vector v = new Vector(values.length);

        for (int i = 0; i < keys.length; i++)
            v.add(values[keys[i]]);

        return v;
    }

   private boolean getIndex(int[] keys, int value)
   {
      for (int i = 0; i < keys.length; i++)
         if (keys[i] == value)
            return true;
      return false;
   }
*/

   public int arrayIndexOf(int[] keys, int value)
   {
      for (int i = 0; i < keys.length; i++)
         if (keys[i] == value)
            return i;
      return -1;
   }

   private int[] createOperatorIndexTable()
   {
      int[] returnValue = new int[HIGHEST_OPCODE];

      // The formats are misnamed - not my fault...
      for (int i = 0; i < oneOpInst.length; i++)
         returnValue[oneOpInst[i]] = FORMAT_THREE;

      for (int i = 0; i < twoOpInst.length; i++)
         returnValue[twoOpInst[i]] = FORMAT_TWO;

      for (int i = 0; i < threeOpInst.length; i++)
         returnValue[threeOpInst[i]] = FORMAT_ONE;

      return returnValue;
   }

   private int generateF1(int opCode, int a, int b, int c)
   {   
      debugWriter.write("F1: ");
      int op = processor.F1(opCode, a, b, c);
      //int op = (opCode << 26) | (a << 21) | (b << 16) | (c & 0xff);
      return op;
   }

   private int generateF2(int opCode, int a, int b, int c)
   {
      debugWriter.write("F2: ");
      int op = processor.F2(opCode, a, b, c);
      // Fix me...
      //int op = (opCode << 26) | (a << 21) | (b << 16) | (c & 0xff);
      return op;
   }

   private int generateF3(int opCode, int c)
   {
      debugWriter.write("F3: ");
      int op = processor.F3(opCode, c);
      //int op = opCode << 26;
      return op;
   }

   public int genOp(int opCode, int a, int b, int c)
   {
      int returnValue = 0;

      debugWriter.write("(" + opCode + " " + a + " " + b + " " + c + ") ");
      //System.out.println("(" + opCode + " " + a + " " + b + " " + c + ") ");

      switch (opTypeIndex[opCode])
      {
      case FORMAT_ONE:
         returnValue = generateF1(opCode, a, b, c);
         break;
      case FORMAT_TWO:
         returnValue = generateF2(opCode, a, b, c);
         break;
      case FORMAT_THREE:
         returnValue = generateF3(opCode, c);
         break;
      default:
         // throw Exception?
      }

      debugWriter.write(opToString(returnValue));
      debugWriter.writeln(" (" + returnValue + ")");
   
      outputInstructions[instructions++] = returnValue;
      return returnValue;
   }

   public String opToString(int op)
   {
      StringBuffer sb = new StringBuffer(32);

      sb.append(Integer.toBinaryString(op));

      while (sb.length() < 32)
         sb.insert(0, "0");

      return sb.toString();
   }

   public int getInstructionCount()
   {
      return instructions;
   }

   public void reverseComparisonInstruction(int instruction)
   {
      processor.disassem(outputInstructions[instruction]);
      int oldOp = arrayIndexOf(comparisonInstructions, processor.op);

      // error?
      if (oldOp == -1)
         return;

      int newOp = comparisonAntiInstructions[oldOp];
      outputInstructions[instruction] = generateF1(newOp,processor.a,processor.b,processor.c);
   }

   public void fixupInstruction(int instruction, int newInst)
   {
      processor.disassem(outputInstructions[instruction]);
      // if we are doing non-F1 instructions and fixups...oops.
      outputInstructions[instruction] = generateF1(processor.op,processor.a,processor.b,newInst-instruction);
   }

   public void writeObjectFile()
   {
      String s;

      PascalUtils.writeInts(objectFilename, outputInstructions, instructions);

      for (int i = 0; i < instructions; i++)
      {
         s = processor.disassemble(outputInstructions[i]);
         assemblerWriter.write(s);
      }

      assemblerWriter.close();
      debugWriter.close();
   }

   public void readObjectFile(String filename)
   {
      outputInstructions = PascalUtils.loadInt(filename, MAX_INST);
      instructions = outputInstructions.length;
   }

   public void executeInstructions()
   {
      int[] temp = new int[instructions];

      System.arraycopy(outputInstructions,0,temp,0,instructions);

      processor.load(temp);

      try
      {  processor.execute();    }
      catch (java.io.IOException ie)
      {  ie.printStackTrace();   }
   }

   /*
   public String testOp(int opCode, int a, int b, int c)
   {
      return opToString(genOp(opCode, a, b, c));
   }
   */

   public static void main(String[] args)
   {
      if (args.length > 0)
      {
         TargetArchitecture ta = new TargetArchitecture();
         ta.readObjectFile(args[0]);
         ta.executeInstructions();
      }
      /*
      TargetArchitecture ta = new TargetArchitecture();

      int[] a = PascalUtils.loadInt(ta.OUTPUT_FILE, ta.MAX_INST);

      for (int i = 0; i < a.length; i++)
         System.out.print(ta.processor.disassemble(a[i]));
      */
      /*
      TargetArchitecture ta = new TargetArchitecture();
      System.out.println("One op:");
      for (int i = 0; i < ta.oneOpInst.length; i++)
         System.out.println(ta.testOp(ta.oneOpInst[i], 1, 2, 3));
      System.out.println("Two op:");
      for (int i = 0; i < ta.twoOpInst.length; i++)
         System.out.println(ta.testOp(ta.twoOpInst[i], 1, 2, 3));
      System.out.println("Three op:");
      for (int i = 0; i < ta.threeOpInst.length; i++)
         System.out.println(ta.testOp(ta.threeOpInst[i], 1, 2, 3));
      */   
   }
}
